import boto3
from boto3.dynamodb.conditions import Key
from botocore.errorfactory import ClientError
from typing import Iterable, Optional

from ..dto import Dto
from ..utils.util import chunk


class ConditionalWriteException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class DynamoDbModel(Dto):
    """
    pkey: A single dynamodb index with column and type as json, such as {'sha256': 'S'}, representing the primary key
    secondary_indices: a list of dynamodb indices with each as json {name: type}
    """

    pkey = dict()
    skey = dict()
    secondary_indices = list()


class DynamoDbTable(object):
    def __init__(
        self, table_name, model, boto_session=None, boto_config=None, endpoint_url=None
    ):
        """
        @param table_name: The name of the table to interact with
        @param model: The DynamoDbModel object to map to
        @param boto_session:
            The boto3 session to use. Will open a session with default region and profile if not provided.
        @param boto_config: botocore config to define communication behavior
        """
        if not issubclass(model, DynamoDbModel):
            raise Exception(
                "Must provide a valid DynamoDbModel class for dynamodb client init."
            )
        if boto_session is None:
            boto_session = boto3.Session()
        self.model = model
        self.pkey_name = list(self.model.pkey.keys())[0]
        self.pkey_type = list(self.model.pkey.values())[0]
        # sort key, optional
        self.skey_name = (
            None if not self.model.skey else list(self.model.skey.keys())[0]
        )
        self.skey_type = (
            None if not self.model.skey else list(self.model.skey.values())[0]
        )
        self.table_name = table_name

        self.client = boto_session.client(
            "dynamodb", config=boto_config, endpoint_url=endpoint_url
        )
        self.table = boto_session.resource(
            "dynamodb", config=boto_config, endpoint_url=endpoint_url
        ).Table(self.table_name)

    def conditional_write(
        self, items, must_be_greater_field, key_ignorelist=None, value_ignorelist=None
    ):
        """
        Conditionally updates DynamoDB Items based on a single field.

        @param items: Items to insert. DTO objects expected. Must have any index keys populated
        @param must_be_greater_field: Name of the field in the DTO object to compare against and existing item,
            if the existing item's attribute is less than the new item, the write will succeed
        @param value_ignorelist: A set of disallowed values. Default includes values that will error on insert in noSQL.
            Keys with these values removed before insert.
        @param key_ignorelist: A set of disallowed keys. Used in the event that you have data used locally that you
            don't want to store.
        """
        if value_ignorelist is None:
            value_ignorelist = {None, ""}
        if key_ignorelist is None:
            key_ignorelist = set()

        for item in items:
            dynamo_item = item.to_dynamo(
                value_ignorelist=value_ignorelist, key_ignorelist=key_ignorelist
            )

            try:
                self.table.put_item(
                    Item=dynamo_item,
                    ConditionExpression=boto3.dynamodb.conditions.Or(
                        boto3.dynamodb.conditions.Attr(
                            list(self.model.pkey.keys())[0]
                        ).not_exists(),
                        boto3.dynamodb.conditions.Or(
                            boto3.dynamodb.conditions.Attr(
                                must_be_greater_field
                            ).not_exists(),
                            boto3.dynamodb.conditions.Attr(must_be_greater_field).lt(
                                dynamo_item[must_be_greater_field]
                            ),
                        ),
                    ),
                )
            except ClientError as e:
                # Ignore Cases where the conditional check failed
                if e.response["Error"]["Code"] != "ConditionalCheckFailedException":
                    raise e
                raise ConditionalWriteException(str(e))

    def write(self, items, value_ignorelist=None, key_ignorelist=None):
        """
        Updates a list of cache items.

        @param items: Items to insert. DTO objects expected. Must have any index keys populated.
        @param value_ignorelist: A set of disallowed values. Default includes values that will error on insert in noSQL.
            Keys with these values removed before insert.
        @param key_ignorelist: A set of disallowed keys. Used in the event that you have data used locally that you
            don't want to store.
        """
        if value_ignorelist is None:
            value_ignorelist = {None, ""}
        if key_ignorelist is None:
            key_ignorelist = set()

        with self.table.batch_writer() as writer:
            for item in items:
                dict_item = item.to_dynamo(
                    value_ignorelist=value_ignorelist, key_ignorelist=key_ignorelist
                )
                writer.put_item(dict_item)

    def delete(self, pkeys: Iterable):
        """
        Deletes a group of items provided their primary key values.


        @param pkeys: A list of the pKeys of the records to delete.
        """
        with self.table.batch_writer() as writer:
            for key in pkeys:
                writer.delete_item(Key={self.pkey_name: key})

    def get(self, keys: Optional[Iterable[tuple]], consistent_read: bool = False):
        """
        Gets a group of items provided their primary key values.

        @param keys: A list of the pKeys of the records to retrieve.
        @param consistent_read: whether or not to perform a strongly consistent read.
        If true, will provide more non-cached results but slow down query.
        @return: A list of the set DTO for the table.
        """
        for requested_chunk in chunk(set(keys), 100):
            if self.skey_name is None:
                keys = [
                    {self.pkey_name: {self.pkey_type: key[0]}}
                    for key in requested_chunk
                ]
            else:
                try:
                    keys = [
                        {
                            self.pkey_name: {self.pkey_type: str(key[0])},
                            self.skey_name: {self.skey_type: str(key[1])},
                        }
                        for key in requested_chunk
                    ]
                except IndexError:
                    raise Exception(
                        f"Sort key {self.skey_name} expected but not provided."
                    )

            response = self.client.batch_get_item(
                RequestItems={
                    self.table_name: {"Keys": keys, "ConsistentRead": consistent_read}
                }
            )

            for item in response["Responses"][self.table_name]:
                yield self.model.from_dynamo(item)

    def query_on_index(self, index: str, column: str, values: Iterable):
        """
        Queries for items on a secondary index given the index name, column name,  and the values to search for

        @param index: the name of the index to search on
        @param column: the name of column to search over
        @param values: a list or set of values to search for
        """
        for value in set(values):
            db_items = self.table.query(
                IndexName=index, KeyConditionExpression=Key(column).eq(value)
            )["Items"]
            for db_item in db_items:
                yield self.model.from_dynamo_dict(db_item, encoded=True)
