# Native imports
import json

# Program imports
from ..utils.util import balanced_chunk
from ..utils.util import balanced_split

# Third-party imports
import boto3


class LambdaDispatcher(object):
    """
    Base class for using the fan-out technique. This lambda is used to
    chunk data and call a bunch of other lambdas.
    """

    def __init__(
        self, lambda_name, chunk_size=None, qualifier=None, desired_num_chunks=None
    ):
        """
        Initialize the class.
        ref: http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.invoke

        @param lambda_name: The name of the lambda method to execute.
        @param chunk_size: The chunk size to use. At most, this many records will be passed to
        the lambdas. Number of chunks minimized while maximizing balanced chunk size.
        @param qualifier: Version or alias of the lambda function. Defaults to the most recent
        version.
        @param desired_num_chunks: Pass this parameter if you want a specific number of balanced chunks,
        instead of maximizing chunk size and using fewer chunks.
        """

        self.lambda_name = lambda_name
        self.chunk_size = chunk_size
        self.qualifier = qualifier
        self.desired_num_chunks = desired_num_chunks

        if self.desired_num_chunks is None and self.chunk_size is None:
            raise Exception("Cannot dispatch without chunk size or number of chunks")

    def dispatch(self, data):
        """
        Partition the data into approximately equal chunks and invoke a lambda event.

        @param data: An iterable containing JSON serializable data.
        """

        client = boto3.client("lambda")
        kwargs = {
            "FunctionName": self.lambda_name,
            "InvocationType": "Event",
            "Payload": None,
        }
        if self.qualifier is not None:
            kwargs["Qualifier"] = self.qualifier

        if self.chunk_size is not None:
            for c in balanced_chunk(data, self.chunk_size, allow_size_increase=False):
                kwargs["Payload"] = json.dumps(c, separators=(",", ":"))
                client.invoke(**kwargs)

        elif self.desired_num_chunks is not None:
            for c in balanced_split(data, self.desired_num_chunks):
                kwargs["Payload"] = json.dumps(c, separators=(",", ":"))
                client.invoke(**kwargs)
