# Native imports
import json
from collections.abc import Hashable

# Program imports
from .aws.dynamodb_utils import dynamo_object_to_dict
from .aws.dynamodb_utils import dict_to_dynamo_dict, dynamo_dict_to_dict


def dto_default(obj, cls, strict=True, encoded=True):
    """
    Creates the specified Dto instance if possible; otherwise, returns None.

    @param obj: The object to convert to a Dto instance.
    @param cls: The Dto class to use.
    @param strict: If True, only existing attributes on the DTO will be added.
    @param encoded: If True, the data is assumed to be encoded, and will first be decoded.
    @return: The Dto instance or None if the value was None.
    """

    return None if obj is None else cls.from_dict(obj, strict, encoded)


def encode_dict(dto, value_ignorelist=None, key_ignorelist=None):
    """
    Encode the Dto object to a dictionary.

    Important note -- if a value in the dto isn't hashable, then the ignorelists will be ignored.

    @param dto: The Dto to encode.
    @param value_ignorelist: A set of values that are invalid and should be removed from the result.
    @param key_ignorelist: A set of keys that are undesired and should be removed from the result.
    """

    if value_ignorelist is None:
        value_ignorelist = set()
    if key_ignorelist is None:
        key_ignorelist = set()

    d = dto.to_dict()
    field_names = dto.encode_names
    for k, v in list(d.items()):
        # Step 1: Perform custom encoding
        encoder = dto.encode_settings.get(k)
        k2 = field_names.get(k, k)
        if encoder is not None:
            v = encoder(v)
        # Delete the original key if an encode_name is used
        if k2 != k:
            del d[k]
        k = k2

        # Step 2: Handle any keys that are to be ignored
        if k in key_ignorelist:
            # Item may have been removed from encoding step
            if k in d:
                del d[k]
            continue

        # Step 3: Recursively process Dtos
        if isinstance(v, Dto):
            d[k] = encode_dict(v, value_ignorelist, key_ignorelist)
        else:
            d[k] = v

        # Step 4: Handle value ignore list
        if isinstance(v, Hashable) and v in value_ignorelist:
            del d[k]

    return d


# noinspection PyCallingNonCallable
def decode_dict(obj, settings=None, names=None, strict=False):
    """
    Decodes a dictionary. This is used to decode JSON dictionaries as well as Python dicts.

    @param obj: The dictionary-like object to decode.
    @param settings: Any decode settings. The key is the key name whose value will be decoded. The
    value is a function that will be used to decode the raw value. It is expected that the function
    takes a single argument, which is the raw value.
    @param names: A dictionary mapping the encoded names to the decoded names. The name replacement
     will occur BEFORE the decode settings are used.
    @param strict: Boolean where if true, we will ignore any keys for which that key name in the
     target dict has a decode name.
    @return: The dictionary with the data decoded.
    """

    d = {}
    if settings is None:
        settings = {}
    if names is None:
        names = {}
    for k, v in obj.items():
        # If strictly adhering to decode rules,
        # only assign to the target dict if the key matching the decode name pair is found.
        # Example: if there is a decode setting {"foo": "bar"} and we have set strict,
        # and "foo" exists in the dict we are decoding - do not use it. Only assign "foo" if "bar" exists.
        if strict and k in names.values():
            continue
        k = names.get(k, k)
        decoder = settings.get(k)
        if decoder is not None:
            d[k] = decoder(v)
        else:
            d[k] = v
    return d


class DtoEncoder(json.JSONEncoder):
    """
    Serializes Dto objects to JSON.
    """

    def default(self, obj):
        if isinstance(obj, Dto):
            return encode_dict(obj)
        return json.JSONEncoder.default(self, obj)


class Dto(object):
    """
    Base class to handle DTO [de]serialization.
    """

    # Mapping of the existing field name to the desired field name. Note that
    # the name change is always performed before decoding and after encoding,
    # such that those settings can always map to the DTO's naming convention.
    encode_names = dict()
    decode_names = dict()

    # Mapping of a field name to a function. The function will be passed the value.
    # The output of the function should be the correctly encoded or decoded value.
    # These attributes are designed to be overwritten by the child.
    encode_settings = dict()
    decode_settings = dict()

    def __str__(self):
        """
        String representation of the class.

        @return: String, representing the class.
        """

        return self.to_json(True)

    def to_json(
        self,
        pretty=False,
        minimize=False,
        key_ignorelist=None,
        ensure_ascii=False,
        jsonl=False,
    ):
        """
        Convert the object to a serialized JSON string.

        @param pretty: If True, the JSON will be pretty printed.
        @param minimize: If you select this option, any keys with null values will be removed.
        @param key_ignorelist: expects a set of keys to remove when putting to json
        @param ensure_ascii: If True, unicode will be stripped, else unicode will be allowed.
        @return: Serialized JSON string.
        """

        if key_ignorelist is None:
            key_ignorelist = set()

        default_kwargs = {"separators": (",", ":")}
        pretty_kwargs = {"sort_keys": True, "indent": 4, "separators": (",", ": ")}

        kwargs = pretty_kwargs if pretty else default_kwargs
        kwargs["ensure_ascii"] = ensure_ascii
        if minimize or jsonl:
            json_str = json.dumps(
                self.to_dict(
                    deep=True,
                    encode=True,
                    value_ignorelist={None},
                    key_ignorelist=key_ignorelist,
                ),
                **kwargs,
            )
            return f"{json_str}\n" if jsonl else json_str
        return json.dumps(self, cls=DtoEncoder, **kwargs)

    def to_dict(
        self, deep=False, encode=False, value_ignorelist=None, key_ignorelist=None
    ):
        """
        Convert the DTO to a dictionary representation.

        Important note -- if a value in the dto isn't hashable, then the ignorelists will be ignored.

        @param deep: If True, all Dto objects will be converted to dictionaries.
        @param encode: If True, the items will be encoded. Note that this parameter ignores deep and will always be a
        deep operation.
        @param value_ignorelist: A set of values that are invalid and should be removed from the result if encoded
        @param key_ignorelist: A set of keys that are undesired and should be removed from the result if encoded
        @return: A dictionary representation of the object.
        """

        if value_ignorelist is None:
            value_ignorelist = set()
        if key_ignorelist is None:
            key_ignorelist = set()

        if encode:
            return encode_dict(
                self, value_ignorelist=value_ignorelist, key_ignorelist=key_ignorelist
            )

        d = dict(vars(self))
        for k, v in list(d.items()):
            if k in key_ignorelist or (
                isinstance(v, Hashable) and v in value_ignorelist
            ):
                del d[k]
                continue
            if deep and isinstance(v, Dto):
                d[k] = v.to_dict(
                    deep=deep,
                    encode=encode,
                    value_ignorelist=value_ignorelist,
                    key_ignorelist=key_ignorelist,
                )
        return d

    def to_dynamo(self, value_ignorelist=None, key_ignorelist=None):
        """
        Convert the DTO to a DynamoDB object.

        @param value_ignorelist: A set of values that are invalid and should be removed from the result if encoded
        @param key_ignorelist: A set of keys that are undesired and should be removed from the result if encoded
        @return: An object that is compatible with DynamoDB.
        """
        if value_ignorelist is None:
            value_ignorelist = set()
        if key_ignorelist is None:
            key_ignorelist = set()

        return dict_to_dynamo_dict(
            self.to_dict(
                deep=True,
                encode=True,
                value_ignorelist=value_ignorelist,
                key_ignorelist=key_ignorelist,
            )
        )

    def update_attributes(self, attributes, strict=False):
        """
        Update and/or add attributes to this DTO.

        @param attributes: An iterable containing key value pairs.
        @param strict: If True, only items that already exist will be updated, i.e. new
        attributes will not be added.
        @return: The updated instance
        """

        a = vars(self)
        for k, v in attributes:
            if strict and k not in a:
                continue
            setattr(self, k, v)
        return self

    @classmethod
    def from_json(cls, serialized_json, strict=True):
        """
        Deserialize the JSON and replace this object's attributes with any matches.

        @param serialized_json: The serialized JSON.
        @param strict: If True, only existing attributes on the DTO will be added.
        @return: The updated instance.
        """

        return cls.from_json_object(
            json.loads(serialized_json), strict=strict, encoded=True
        )

    @classmethod
    def from_dynamo_dict(cls, dict_obj, strict=True, encoded=False):
        return cls.from_dict(dynamo_dict_to_dict(dict_obj), strict, encoded)

    @classmethod
    def from_dict(cls, dict_obj, strict=True, encoded=False):
        """
        Replace this object's attributes with the items in the dictionary-like object.

        @param dict_obj: The dictionary.
        @param strict: If True, only existing attributes on the DTO will be added.
        @param encoded: If True, the data is assumed to be encoded, and will first be decoded.
        @return: The updated instance.
        """

        # Update attributes using the direct values of the dictionary
        if not encoded:
            return cls().update_attributes(dict_obj.items(), strict)

        # Update the attributes after decoding
        return cls().update_attributes(
            decode_dict(
                dict_obj, cls.decode_settings, cls.decode_names, strict=strict
            ).items(),
            strict,
        )

    @classmethod
    def from_json_object(cls, obj, *args, **kwargs):
        return cls.from_dict(obj, *args, **kwargs)

    @classmethod
    def from_dynamo(cls, obj):
        """
        Convert the DynamoDB object to the DTO.

        @param obj: The DynamoDB object.
        @return: The DTO.
        """

        return cls.from_dict(dynamo_object_to_dict(obj), strict=True, encoded=True)
