import cProfile
import itertools
import json
import os
import re
import signal
import threading
from time import sleep, time
import traceback
import uuid
from abc import ABC, abstractmethod

import flask
from flask_swagger import swagger
import requests
from boto3 import Session
from flask import Flask, jsonify, request, g, abort
from flask_principal import Principal, Identity, RoleNeed, Need, Permission
from itsdangerous import TimedJSONWebSignatureSerializer
from werkzeug.datastructures import Headers

from .tic_status import tic_status
from .base_config import BaseConfig
from .statsd_aggregator import StatsDAggregator
import tic.shared.utils.kubernetes_util as kubernetes_util
from tic.auth.auth_util import AuthUtil
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.configuration import configuration
from tic.shared.aws.kms_client import KmsClient
from tic.shared.api_doc_settings import ApiDocSettings


def do_cprofile(func):
    def profiled_func(*args, **kwargs):
        profile = cProfile.Profile()
        try:
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            return result
        finally:
            profile.print_stats()

    return profiled_func


# region Standard permissions
# The identity must have at least one of the specified permissions
use_header_cust_token_permission = Permission(Need("read", "header_customer_token"))
download_binary_permission = Permission(Need("read", "binary"))
upload_binary_permission = Permission(Need("create", "binary"))
read_binary_metadata_permission = Permission(Need("read", "binary_metadata"))
# endregion

# cbfeed/mgmt permissions:
create_feed_data_permission = Permission(Need("create", "feed_data"))
read_feed_data_permission = Permission(Need("read", "feed_data"))
update_feed_data_permission = Permission(Need("update", "feed_data"))
delete_feed_data_permission = Permission(Need("delete", "feed_data"))

# cb_server_module permissions:
create_cb_server_records = Permission(Need("create", "cb_server_records"))

# customer_data permissions
create_customer_record_permission = Permission(Need("create", "customer_record"))
read_customer_record_permission = Permission(Need("read", "customer_record"))
update_customer_record_permission = Permission(Need("update", "customer_record"))
delete_customer_record_permission = Permission(Need("delete", "customer_record"))

# feedback permissions
create_feedback_permission = Permission(Need("create", "feedback"))
read_feedback_permission = Permission(Need("read", "feedback"))
update_feedback_permission = Permission(Need("update", "feedback"))
delete_feedback_permission = Permission(Need("delete", "feedback"))

# defense permissions
create_defense_data_permission = Permission(Need("create", "defense_data"))
read_defense_data_permission = Permission(Need("read", "defense_data"))
update_defense_data_permission = Permission(Need("update", "defense_data"))
delete_defense_data_permission = Permission(Need("delete", "defense_data"))

# service status permissions
read_service_status_permission = Permission(Need("read", "service_status"))

# reputation permissions
create_reputation_data_permission = Permission(Need("create", "reputation_data"))
read_reputation_data_permission = Permission(Need("read", "reputation_data"))
update_reputation_data_permission = Permission(Need("update", "reputation_data"))
delete_reputation_data_permission = Permission(Need("delete", "reputation_data"))

# kpi permissions
create_kpi_data_permission = Permission(Need("create", "kpi_data"))
read_kpi_data_permission = Permission(Need("read", "kpi_data"))
update_kpi_data_permission = Permission(Need("update", "kpi_data"))
delete_kpi_data_permission = Permission(Need("delete", "kpi_data"))

# UBS permissions
create_ubs_data_permission = Permission(Need("create", "ubs_data"))
read_ubs_data_permission = Permission(Need("read", "ubs_data"))
update_ubs_data_permission = Permission(Need("update", "ubs_data"))
delete_ubs_data_permission = Permission(Need("delete", "ubs_data"))

# upload permissions
create_upload_data_permission = Permission(Need("create", "upload_data"))
read_upload_data_permission = Permission(Need("read", "upload_data"))
update_upload_data_permission = Permission(Need("update", "upload_data"))
delete_upload_data_permission = Permission(Need("delete", "upload_data"))

# FQP permissions for admin/analyst role
read_fqp_data = Permission(Need("read", "fqp_data"))
update_fqp_data = Permission(
    Need("update", "fqp_data")
)  # no distinction between create and update for FQP server

# FQP permissions for customer FQP client
read_fqp_query_jobs = Permission(Need("read", "fqp_query_jobs"))
update_fqp_query_jobs = Permission(Need("update", "fqp_query_jobs"))

# Trust/Threat Score Permissions. Used to interact with trust/threat scores directly.
create_trust_threat_scores = Permission(Need("create", "trust_threat_scores"))
read_trust_threat_scores = Permission(Need("read", "trust_threat_scores"))
update_trust_threat_scores = Permission(Need("update", "trust_threat_scores"))
delete_trust_threat_scores = Permission(Need("delete", "trust_threat_scores"))

# Defense Rep Score Permissions. Used to interact with trust/threat scores by providing Defense Reputation objects.
create_defense_rep_scores = Permission(Need("create", "defense_rep_scores"))
read_defense_rep_scores = Permission(Need("read", "defense_rep_scores"))
update_defense_rep_scores = Permission(Need("update", "defense_rep_scores"))
delete_defense_rep_scores = Permission(Need("delete", "defense_rep_scores"))

# avira_updater permissions
update_avira_updater_permission = Permission(Need("update", "avira_updater"))

# config data read permission
read_config_data_permission = Permission(Need("read", "config_data"))


class BasePrincipal(ABC):
    @abstractmethod
    def create_principal(self, app):
        pass


class NoHealthyInstancesException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class AuthException(Exception):
    pass


class TICConfiguration(BaseConfig):
    """
    Base TIC services configuration class

    TIC Configuration is stored in ( sorted by load priority ):
        1. Hardcoded
        2. Environment variable
        3. Configuration file

    Each configuration layer is overridden by the next (e.g. hardcoded values are overridden by env values )

    get(<key>) method is used to fetch configuration values. It will return the "base" config value if a "service" value
    was not specified.

    Override load_config_from_xxx in your own configuration class if needed
    """

    def __init__(
        self,
        service_name,
        port,
        statsd_prefix=None,
        statsd_aggregation_time: int = 0,
        api_docs_settings: ApiDocSettings = None,
    ):
        super().pre_init(service_name)  # This is gross, but solves some pain points
        self.baseConfig = {}
        self.config = {}
        self.configPrefix = self.__class__.__name__ + "/"
        self._base_config_prefix = "TICBaseConfig/"
        self.load_config()
        super().__init__(
            service_name,
            port,
            statsd_prefix,
            statsd_aggregation_time,
            api_docs_settings,
            True,
        )

        # We only have an ELB for certificatevalidation since it is a service running on Windows
        self.elb_only_services = ["certificatevalidation"]

        self.init_boto_session()

        # Assert that the configs are not empty before we say config was successfully loaded.
        # TODO: Technically this shouldn't get set here but instead _after_ the full config is set.
        #       This will need to be addressed as we migrate forwards.
        if self.config and self.baseConfig:
            self.loaded = True

    def load_core(self):
        self.aws_region = self.get("AWS", "aws_region")
        self.logging_level = self.get("logging", "log_level", default="INFO")
        self.statsd_host = self.get("metrics", "statsd_host")
        self.statsd_port = int(self.get("metrics", "statsd_port", default=8125))

    def init_boto_session(self):
        boto_session = Session(region_name=self.aws_region)
        total_seconds = 0
        num_retries = 0

        # retry 25s- k8s waits 30 before calling first ready_check, 60 before first health_check
        while total_seconds < 25:
            try:
                if boto_session.get_credentials() is not None:
                    self.boto_session = boto_session
                    break
            except Exception:
                num_retries += 1

            wait_time = 2**num_retries
            total_seconds += wait_time
            sleep(wait_time)

    def load_config(self):
        # By default we load the config from files in on boot
        self.baseConfig = configuration.load_config_from_files(
            self.tic_environment, root_key="TICBaseConfig"
        )
        self.config = configuration.load_config_from_files(
            self.tic_environment, root_key=self.raw_component_name
        )

    def config_changed(self):
        pass

    def get_healthy_service_instance(self, service_name):
        """
        :param service_name: Service name that you want the know the healthy node for.
        :return: None if there is an error, service node information is successful
        """

        return self.get_elb_service(service_name)

    def get_healthy_service_uri(self, service_name, api_version, *args):
        service_instance = self.get_elb_service(service_name)

        if service_instance:
            service_uri = "http://{0}:{1}/{2}/".format(
                service_instance["node"]["Address"],
                service_instance["service"]["Port"],
                api_version,
            ) + "/".join(part for part in args)
            return service_uri

    # Get an ELB for the service if it exists
    def get_elb_service(self, service_name):
        service_name_lower = service_name.lower()
        if kubernetes_util.KUBERNETES_ENABLED:
            # If Running in kubernetes use kubernets load balancer name.
            if service_name_lower in kubernetes_util.SERVICE_TRANSLATION:
                return {
                    "node": {
                        "Address": kubernetes_util.SERVICE_TRANSLATION[
                            service_name_lower
                        ]
                        + self.ci_postfix
                    },
                    "service": {"Port": "10000"},
                }
            else:
                if service_name_lower not in self.elb_only_services:
                    self.logger.debug(
                        f"ELB for {service_name_lower} not found. "
                        "No entry in SERVICE_TRANSLATION table."
                    )

        config_service = self.get("Services", service_name)
        if config_service:
            sections = config_service.split(":", 1)
            service_address = sections[0]
            port = int(sections[1]) if len(sections) > 1 else 80
            return {"node": {"Address": service_address}, "service": {"Port": port}}

        raise NoHealthyInstancesException(
            f"Unable to get a healthy instance of {service_name}"
        )

    def get_section(self, config_section):
        root_key = self.raw_component_name
        service_configs = self.config.get(root_key, {})
        base_configs = self.baseConfig.get("TICBaseConfig", {})
        return service_configs.get(config_section) or base_configs.get(config_section)

    # Note: Method signatures don't match. This is known, but done to avoid breaking changes.
    def get(self, config_section, config_key, default=None):
        root_key = self.raw_component_name
        service_configs = self.config.get(root_key, {})
        service_config_section = service_configs.get(config_section, {})

        base_configs = self.baseConfig.get("TICBaseConfig", {})
        base_config_section = base_configs.get(config_section, {})

        return (
            service_config_section.get(config_key)
            or base_config_section.get(config_key)
            or default
        )


class SQSQueue(object):
    def __init__(
        self,
        queue_name,
        queue_region,
        config,
        num_messages=1,
        attributes=None,
        access_key_id=None,
        access_key=None,
        sqs_endpoint_url=None,
    ):
        self.logger = config.logger
        self.statsd = config.statsd
        self.config = config
        # boto_session should assume credentials if configured
        self.boto_session = (
            self.config.boto_session
            if self.config.boto_session is not None
            else Session(region_name=queue_region)
        )

        if access_key_id and access_key:
            self.process_thread_boto_session = Session(
                aws_access_key_id=access_key_id,
                aws_secret_access_key=access_key,
                region_name=queue_region,
            )
        else:
            self.process_thread_boto_session = Session(region_name=queue_region)

        self.sqs_client = SqsClient(
            boto_session=self.boto_session, endpoint_url=sqs_endpoint_url
        )

        self.queue_url = self.sqs_client.get_queue_url(queue_name)
        self.queue_region = queue_region
        self.get_healthy_service_uri = self.config.get_healthy_service_uri
        self.get_healthy_service_instance = self.config.get_healthy_service_instance
        self.listen_enabled = True
        self.num_messages = num_messages
        self.attributes = attributes
        self.first_boot = True

        self.logger_extra = {"message_id": None}

        self._start_sqs_loop()

    def post(self):
        if self.listen_enabled:
            return 200
        self.listen_enabled = True
        return 200

    def delete(self):
        self.listen_enabled = False
        self.sqs_loop_thread.join()
        self.logger.debug("SQS Listen stopped")
        return 200

    def get(self):
        return {"ListenStatus": str(self.listen_enabled)}, 200

    def _start_sqs_loop(self):
        self.logger.debug("Creating SQS Listen thread")
        self.sqs_loop_thread = threading.Thread(target=self.sqs_loop)
        self.sqs_loop_thread.daemon = True
        self.sqs_loop_thread.start()

    def sqs_loop(self):
        while True:
            try:
                self.logger.debug("SQS Listen thread started")

                tic_status.exit = False

                while self.listen_enabled and not tic_status.killed:
                    # Sleep is enabled - don't read any messages from the queue.
                    if self.check_sleep():
                        continue

                    messages = self.sqs_client.receive_message(
                        queue=self.queue_url,
                        num_messages=self.num_messages,
                        wait_seconds=20,
                        attributes=self.attributes,
                    )

                    for message in messages.get("Messages", list()):
                        self.process_message(message)

                self.logger.warning("Exiting SQS Listen thread")
                tic_status.exit = True
                if tic_status.killed is True:
                    os._exit(0)

            except Exception:
                self.logger.exception(
                    "sqs_loop: exception encountered, waiting and retrying sqs connection"
                )
                sleep(10)

    def check_sleep(self):
        """
        Function to override by implementing classes. If this function returns True, messages won't be read from the queue.

        sleep should be called in the implementing class.
        """
        # sleep 5 seconds on first boot to allow time to finish establishing clients & connections.
        # otherwise the process thread attempts to process immediately
        if self.first_boot:
            sleep(5)
            self.first_boot = False

    def process_message(self, message):
        pass


class TicPrincipal(BasePrincipal):
    def __init__(self, config, allow_anonymous=False):
        self.config = config
        self.logger = config.logger
        self.app = None
        self.allow_anonymous = allow_anonymous
        self.kms_client = KmsClient(self.config.boto_session)
        self.auth_session = requests.Session()
        self.principal = None
        self.auth_util = None
        self.auth_token_key = None
        self.auth_headers = None

    def create_principal(self, app):
        try:
            self.app = app
            self.principal = Principal(self.app, use_sessions=False)
            self.principal.identity_loader(self.authenticate)
            self.auth_util = AuthUtil(
                logger=self.logger, boto_session=self.config.boto_session
            )

            start_time = time()

            self.auth_token_key = self.kms_client.decrypt_string(
                self.config.get("auth", "encrypted_token_key")
            )

            complete = start_time - time()

            if complete > int(
                self.config.get("monitoring", "create_principal_decrypt_timeout", 3)
            ):
                self.logger.warning(
                    f"Timeout exceeded on kms_client.decrypt_string - {complete}"
                )

            self.auth_headers = Headers()

            return self.principal
        except Exception as e:
            self.logger.exception("Failed to create principal.")
            raise e

    def retrieve_customer(
        self, customer_token=None, server_token=None, impersonate_customer_token=None
    ):
        if server_token is not None:
            query_params = (
                f"?customer-token={customer_token}&server-token={server_token}"
            )
        else:
            query_params = f"?customer-token={customer_token}"

        if impersonate_customer_token:
            query_params += f"&impersonate-customer-token={impersonate_customer_token}"

        auth_url = self.config.get_healthy_service_uri(
            "auth", "v1", "customer-auth-token"
        )
        if auth_url is None:
            raise Exception("Failed to get auth url")

        # get auth token from auth service
        response = self.auth_session.get(
            auth_url + query_params, headers=self.auth_headers
        )

        # setup identity, etc
        if response.status_code == 200:
            response_data = json.loads(response.content)
            return response_data
        elif response.status_code == 404:
            return None
        else:
            raise AuthException(
                f"Call to auth service failed with {response.status_code=}"
            )

    # @do_cprofile
    def authenticate(self):
        """

        Handle the following cases:
        (1) Request header contains auth_token - has already been authenticated
        (2) Request header contains certificate name i.e. call has been made by cb server / alliance
            Request may also contain server token
        (3) Request query parameters contains user_key and user_secret i.e. call from ticadmin or TR ...

        If authentication succeeds:
            Create an Identity with the appropriate role, permission data

        Impersonation notes:
            customer_token = caller (Alliance)
            impersonate_customer_token = customer (Company 1)
            raise_impersonation_failure = false

            Impersonation will succeed if:
                The caller (Alliance) has the permission to impersonate. This permission allows caller to impersonate anybody.
                AND
                The customer (Company 1) was found and is valid in the auth db.

            If impersonation fails the values of the raise_impersonation_failure flag will determine
            if we need to raise an error. In most cases this flag is set to false.


        Logging requirements:
            Log one statement for the successful authentication case and one statement for the failure case.

        """
        start_time = time()
        time_splits = []

        # ignore health_checks
        if (
            request.endpoint == "health_check"
            or request.endpoint == "ready_check"
            or request.endpoint == ""
        ):
            return

        # with PyCallGraph(output=self.graphviz):
        # define what can be anonymous
        allow_anonymous = self.allow_anonymous or request.endpoint == "catch_all_ui"
        customer_token = None
        server_token = None
        impersonate_customer_token = None
        user_key = None

        try:
            user_key = request.args.get("user-key") or request.headers.get(
                "X-Tic-User-Key"
            )
            user_secret = request.args.get("user-secret") or request.headers.get(
                "X-Tic-User-Secret"
            )

            self.auth_headers["request_id"] = _get_request_id()

            if "auth_token" in request.headers:
                # setup identity, etc
                self.process_auth_token(request.headers["auth_token"])
                time_splits.append({"process_auth_token": time() - start_time})
                self.logger.debug(
                    f"Successfully authenticated {request.url=} using auth token in header"
                )

            elif user_key is not None and user_secret is not None:

                self.logger.debug(f"{user_key=}, {user_secret=}")

                auth_url = self.config.get_healthy_service_uri(
                    "auth", "v1", "user-auth-token"
                )
                if auth_url is None:
                    raise Exception("Failed to get auth url")

                query_params = "?user-key=" + user_key + "&user-secret=" + user_secret
                response = self.auth_session.get(
                    auth_url + query_params, headers=self.auth_headers
                )

                time_splits.append({"request_auth": time() - start_time})

                # setup identity, etc
                if response.status_code == 200:
                    response_data = json.loads(response.content)
                    self.process_auth_token(response_data["token"])

                else:
                    raise AuthException(
                        f"Call to auth service failed with {response.status_code=}"
                    )

                self.logger.debug(f"Successfully authenticated {request.url=}")

            elif (
                "X-Ssl-Dn" in request.headers or "X-Cb-CustomerToken" in request.headers
            ):
                time_splits.append(
                    {"x-ssl-dn_or_x-cb-customertoken_start": time() - start_time}
                )
                # get cert_cn from header
                customer_token, header_token = lookup_customer_token(request)
                server_token = lookup_server_token(request)
                impersonate_customer_token = request.headers.get(
                    "X-Cb-ImpersonateCustomer"
                )
                raise_impersonation_failure = request.headers.get(
                    "X-Cb-RaiseImpersonationFailure", "False"
                )
                raise_impersonation_failure = raise_impersonation_failure.lower() in [
                    "true",
                    "1",
                    "yes",
                ]

                if customer_token is None:
                    raise AuthException("Missing customer token")

                impersonation_succeeded = None
                # if request is for an impersonated call and not using the bypass token, attempt to impersonate
                if impersonate_customer_token and not header_token:
                    time_splits.append({"pre_retrieve_customer": time() - start_time})
                    customer_data = self.retrieve_customer(
                        customer_token=customer_token,
                        server_token=server_token,
                        impersonate_customer_token=impersonate_customer_token,
                    )

                    impersonation_succeeded = True

                    time_splits.append(
                        {
                            "impersonate_customer_token_retrieve_customer": time()
                            - start_time
                        }
                    )

                    # If they submitted an impersonate customer header, but that token is not valid,
                    # downgrade to just authenticating original caller UNless raise impersonation failures is set.
                    if not customer_data and not raise_impersonation_failure:
                        customer_data = self.retrieve_customer(
                            customer_token=customer_token, server_token=server_token
                        )

                        impersonation_succeeded = False

                        time_splits.append(
                            {"not_customer_data_retrieve_customer": time() - start_time}
                        )

                # If we have successful ClientCertCheck and it is a Cb Defense Customer - bypass auth
                elif (
                    customer_token.startswith("cb_defense_")
                    or customer_token.startswith("cb_reputation_")
                ) and request.headers.get("X-Cb-ClientCertCheck") == "SUCCESS":
                    customer_data = self.retrieve_reputation_customer(customer_token)
                    self.logger.debug(
                        f"Successfully authenticate {customer_token} CbD SSL Cert bypassing auth"
                    )
                # If we have successful ClientCertCheck and it is a Cb Defense Customer - bypass auth
                elif (
                    customer_token.startswith("cb_ubs_")
                    and request.headers.get("X-Cb-ClientCertCheck") == "SUCCESS"
                ):
                    customer_data = self.retrieve_ubs_customer(customer_token)
                    self.logger.debug(
                        f"Successfully authenticate {customer_token} UBS SSL Cert bypassing auth"
                    )
                else:
                    time_splits.append({"pre_retrieve_customer": time() - start_time})

                    customer_data = self.retrieve_customer(
                        customer_token=customer_token, server_token=server_token
                    )

                    time_splits.append({"retrieve_customer": time() - start_time})
                if not customer_data:
                    raise AuthException(f"Unable to retrieve customer {customer_token}")

                identity = self.process_auth_token(customer_data["token"])
                if header_token and not identity.can(use_header_cust_token_permission):
                    raise AuthException("Unable to authenticate")

                self.logger.debug(
                    f"Successfully authenticated {request.url=} "
                    f"{customer_token=} "
                    f"{server_token=} "
                    f"{impersonate_customer_token=} "
                    f"{impersonation_succeeded=} "
                    f"{request.headers=}"
                )

            elif "X-Tic-Activation-Token" in request.headers:
                activation_token = request.headers.get("X-Tic-Activation-Token", None)
                self.logger.debug(f"Attempting to authenticate {activation_token=}")

                auth_url = self.config.get_healthy_service_uri(
                    "auth", "v1", "activation-token-auth-token/"
                )
                if auth_url is None:
                    raise Exception("Failed to get auth url")

                response = self.auth_session.get(
                    auth_url + activation_token, headers=self.auth_headers
                )

                time_splits.append({"request_auth": time() - start_time})

                # setup identity, etc
                if response.status_code == 200:
                    response_data = json.loads(response.content)
                    self.process_auth_token(response_data["token"])

                else:
                    raise AuthException(
                        f"Call to auth service failed with {response.status_code=}"
                    )

                self.logger.debug(
                    f"Successfully authenticated {request.url=} {activation_token=} {request.headers=}"
                )

            elif "access_key_id" in request.headers:
                raise NotImplementedError(
                    "Cannot handle access_key_id and secret_access_key yet."
                )

            else:
                if allow_anonymous:
                    return
                else:
                    raise AuthException(f"Missing authentication data. {request.url=}")

        except Exception as e:
            self.logger.exception(
                f"Authentication failed for {request.url=} "
                f"{customer_token=} "
                f"{impersonate_customer_token=}"
                f"{server_token=}"
                f"{user_key=}"
                f"{request.headers=}"
            )
            raise e

        finally:
            if time() - start_time > int(
                self.config.get("monitoring", "auth_timeout", 3)
            ):
                self.logger.warning(
                    f"Timeout exceeded on authenticate for {request.url=} "
                    f"{customer_token=} "
                    f"{impersonate_customer_token=} "
                    f"{server_token=} "
                    f"{user_key=} "
                    f"{request.headers=}"
                )

                self.logger.warning(
                    "Timeout exceeded on authenticate\n{0}".format(
                        "\n".join([str(split) for split in time_splits])
                    )
                )

    def retrieve_ubs_customer(self, customer_token):
        user = {
            "customer_token": customer_token,
            "credentials": {
                "name": customer_token,
                "roles": [
                    {
                        "name": "cb-customer",
                        "permissions": [
                            {
                                "action": {"name": "read"},
                                "resource": {"name": "ubs_data"},
                            }
                        ],
                    }
                ],
                "enabled": True,
                "customer_token": customer_token,
                "id": 4000,
                "permissions": [],
            },
            "user_id": 4000,
            "impersonated_by_token": None,
            "server_token": None,
        }

        serializer = TimedJSONWebSignatureSerializer(
            self.auth_token_key, expires_in=43200 + int(60) * 60, algorithm_name="HS256"
        )

        token = serializer.dumps(user)

        return {"token": token}

    def retrieve_reputation_customer(self, customer_token):
        """
        Return the auth token for a Cb Defense customer.
        """

        user = {
            "customer_token": customer_token,
            "credentials": {
                "name": customer_token,
                "roles": [
                    {
                        "name": "cb-customer",
                        "permissions": [
                            {
                                "action": {"name": "read"},
                                "resource": {"name": "defense_data"},
                            },
                            {
                                "action": {"name": "read"},
                                "resource": {"name": "reputation_data"},
                            },
                            {
                                "action": {"name": "create"},
                                "resource": {"name": "ubs_data"},
                            },
                        ],
                    }
                ],
                "enabled": True,
                "customer_token": customer_token,
                "id": 4000,
                "permissions": [],
            },
            "user_id": 4000,
            "impersonated_by_token": None,
            "server_token": None,
        }

        serializer = TimedJSONWebSignatureSerializer(
            self.auth_token_key, expires_in=43200 + int(60) * 60, algorithm_name="HS256"
        )

        token = serializer.dumps(user)

        return {"token": token}

    def process_auth_token(self, auth_token):
        auth_token_data = self.auth_util.get_auth_token_data(
            auth_token, self.auth_token_key
        )
        identity = self.get_identity(auth_token_data)
        self.principal.set_identity(identity)

        tic_principal_data = TicPrincipalData()
        tic_principal_data.user_id = auth_token_data["user_id"]
        tic_principal_data.user_name = auth_token_data["credentials"]["name"]
        tic_principal_data.customer_token = auth_token_data["customer_token"]
        tic_principal_data.server_token = auth_token_data["server_token"]
        tic_principal_data.auth_token = auth_token
        tic_principal_data.auth_token_key = self.auth_token_key
        g.tic_principal_data = tic_principal_data

        return identity

    def get_identity(self, auth_token_data):
        """
        Decrypt the specified auth_token and create an identity using the data in the auth token
        :param auth_token_data:
        :return:
        """
        identity = Identity(auth_token_data["user_id"])

        # populate Needs
        credentials = auth_token_data["credentials"]

        for role in credentials["roles"]:
            identity.provides.add(RoleNeed(role["name"]))

            # role permissions
            for permission in role["permissions"]:
                identity.provides.add(
                    Need(permission["action"]["name"], permission["resource"]["name"])
                )

        # user permissions
        for permission in credentials["permissions"]:
            identity.provides.add(
                Need(permission["action"]["name"], permission["resource"]["name"])
            )

        return identity


class TicPrincipalData(object):
    def __init__(self):
        self.user_id = None
        self.customer_token = None
        self.server_token = None
        self.auth_token = None
        self.auth_token_key = None
        self.user_name = None


class TICService(object):
    access_key_id = None
    secret_access_key = None

    def __init__(self, config, principal=None):
        self.app = Flask(__name__)

        if config == "swagger":
            return

        self.app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

        self.config = config
        self.logger = config.logger
        self.statsd = config.statsd
        self.statsd_tcp = config.statsd_tcp
        self.boto_session = config.boto_session
        signal.signal(signal.SIGTERM, self.sigterm_handler)

        self.app.add_url_rule(
            "/configuration", "configuration", self.configuration, methods=["GET"]
        )
        self.app.add_url_rule("/health_check", "health_check", self.health_check)
        self.app.add_url_rule("/ready_check", "ready_check", self.ready_check)
        self.app.add_url_rule("/v1/health_check", "health_check", self.health_check)
        self.app.add_url_rule("/v1_health_check", "health_check", self.health_check)
        self.app.add_url_rule("/v1/ready_check", "ready_check", self.ready_check)
        self.app.add_url_rule("/whoami", "whoami", self.whoami)
        self.app.add_url_rule(
            "/v1/info/api_docs", "api_docs", self.api_docs, methods=["GET"]
        )
        self.app.add_url_rule("/", "", self.index_request)

        self.app.register_error_handler(500, self._server_error_handler)

        if principal:
            try:
                self.principal = principal.create_principal(self.app)
            except Exception as e:
                self.logger.exception("Failed to create principal.")
                raise e

        self.ip = self.get_local_ip()

        @self.app.before_request
        def before_request():
            # Nothing to do if we have a request_id in g
            if getattr(g, "request_id", None) is None:
                # if we have a request_id in the header - store it in g
                request_id = request.headers.get("request_id")

                if not request_id:
                    request_id = request.headers.get("X-Request-ID")

                g.request_id = request_id or str(uuid.uuid4())

        @self.app.after_request
        def after_request(response):
            # Set the Request ID in the header if it isn't set.
            if "X-Request-ID" not in response.headers:
                response.headers["X-Request-ID"] = _get_request_id()

            return response

    def sigterm_handler(self, signal, frame):
        tic_status.killed = True  # gracefully stop current operations
        tic_status.ready = False  # process no longer ready
        if isinstance(self.statsd, StatsDAggregator):
            self.statsd.stop_metrics_loop()
            self.statsd_tcp.stop_metrics_loop()
        self.logger.debug("SIGTERM Received! Goodbye cruel world...")

    @property
    def tic_principal(self):
        return getattr(g, "tic_principal_data", None)

    def _server_error_handler(self, e):
        reason = "internal server error"
        status_code = 500
        caller_context = ""

        if hasattr(self, "tic_principal"):
            if self.tic_principal:
                caller_context = f"{self.tic_principal.customer_token}:{self.tic_principal.server_token}"

        # flask 1.1 returns InternalServerError for all exceptions at the root - use e.original_exception rather than e.
        # https://palletsprojects.com/blog/flask-1-1-released/
        if hasattr(e, "original_exception") and isinstance(
            e.original_exception, AuthException
        ):
            reason = "Not allowed"
            status_code = 401
        else:
            self.logger.exception(
                f"Unexpected error: {str(e)}\nfrom: {caller_context}\n{traceback.format_exc()}"
            )

        self.logger.exception(
            f"{request.method} call to {request.url=} for {caller_context=} failed with {status_code=}."
        )

        resp = jsonify({"reason": reason})
        resp.status_code = status_code
        resp.headers["X-Request-ID"] = _get_request_id()

        return resp

    @staticmethod
    def index_request():
        """
        Return nothing when the index is requested.
        """
        return jsonify({}), 200

    def health_check(self):
        """
        Used by Kubernetes to determine service health

        :statuscode 200: Healthy
        :statuscode 500: Not Healthy
        """
        if self.boto_session is not None:
            try:
                if self.boto_session.get_credentials() is None:
                    raise Exception("Boto session credentials None in health_check.")
                if self.boto_session.get_credentials().secret_key is None:
                    raise Exception("Boto session Secret Key None in health_check.")
            except Exception as e:
                tic_status.healthy = False
                self.logger.exception(f"Health Check Exception!: {str(e)}")
                return jsonify({"health": "ERROR"}), 500
        else:
            self.logger.info("Health Check Success: No Valid Boto Session")
            return jsonify({"health": "OK"}), 200
        self.logger.info("Health Check Success!")
        return jsonify({"health": "OK"}), 200

    def ready_check(self):
        """
        Used by Kubernetes in the readiness probe

        :statuscode 200: Ready
        :statuscode 503: Not Ready
        """
        if self.boto_session is not None:
            try:
                if self.boto_session.get_credentials() is None:
                    raise Exception("Boto session credentials None in ready_check.")
            except Exception as e:
                tic_status.ready = False
                self.logger.exception(str(e))
        if tic_status.ready is True and tic_status.config_loaded is True:
            return jsonify({"ready": "True"}), 200
        return jsonify({"ready": "False"}), 503

    def whoami(self):
        """
        Used by any client to determine who they are authenticated as

        :statuscode 200: Healthy
        """

        return jsonify({"customer_token": self.tic_principal.customer_token}), 200

    def api_docs(self):
        template = {
            "swagger": "2.0",
            "info": {
                "title": self.config.api_docs_settings.external_service_name,
                "description": f"API for the {self.config.api_docs_settings.external_service_name}",
                "contact": {"email": "holdmybeer@carbonblack.com"},
                "version": self.config.api_docs_settings.semver_api_version,
            },
            "basePath": f"/{self.config.api_docs_settings.base_path}",
        }
        return jsonify(swagger(self.app, template=template))

    @read_config_data_permission.require()
    def configuration(self):
        """
        Used to respond to queries about currently set configuration(s)
        """
        return (
            jsonify(
                {"config": self.config.config, "baseConfig": self.config.baseConfig}
            ),
            200,
        )

    @staticmethod
    def not_found():
        abort(404)

    def lookup_customer_token(self, request):
        customer_token, header_token = lookup_customer_token(request)
        if header_token:
            self.logger.warning(f"Header token: {customer_token}")
        return customer_token

    def lookup_server_token(self, request):
        server_token = lookup_server_token(request)
        return server_token

    def get_local_ip(self):
        ip = "Unknown"
        try:
            # try aws
            # See: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
            response = requests.get(
                "http://169.254.169.254/latest/meta-data/local-ipv4", timeout=1
            )
            if response.status_code == 200:
                ip = response.text
        except Exception:
            self.logger.debug("Failed to get ip address from aws metadata")
            # http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib?lq=1
            # Don't want to get too fancy about this - because this is not the main path.
            # ip = socket.gethostbyname(socket.gethostname())

        # replace '.' with '-'
        return ip.replace(".", "-")

    def get_pass_through_header(self):
        headers = {"pass_through": "1"}

        if hasattr(request, "headers"):
            cust_token_header = request.headers.get("X-Ssl-Dn")
            if cust_token_header:
                headers["X-Ssl-Dn"] = cust_token_header
            temp_activation_token = request.headers.get("X-Tic-Activation-Token")
            if temp_activation_token:
                headers["X-Tic-Activation-Token"] = temp_activation_token

        headers["auth_token"] = self.tic_principal.auth_token

        # TODO: using json by default - revisit
        headers["Content-Type"] = "application/json"

        request_id = _get_request_id()
        if request_id:
            headers["request_id"] = request_id

        return headers

    def has_permission(self, action_name, resource_name):
        # TODO why initialize here.
        auth_util = AuthUtil(self.logger, self.boto_session)
        auth_data = auth_util.get_auth_token_data(
            self.tic_principal.auth_token, self.tic_principal.auth_token_key
        )

        # Permissions are at the top level (user permissions) and at the level of the role (role permissions).
        # Must account for both
        permissions = auth_data["credentials"]["permissions"] + list(
            itertools.chain.from_iterable(
                [x["permissions"] for x in auth_data["credentials"]["roles"]]
            )
        )

        for permission in permissions:
            if (
                permission["action"]["name"] == action_name
                and permission["resource"]["name"] == resource_name
            ):
                return True

        return False


def lookup_server_token(rq):
    return rq.headers.get("X-Cb-ServerToken", None)


def lookup_customer_token(rq):
    header_token = False
    customer_token = parse_customer_token(rq.headers.get("X-Ssl-Dn", None))
    if not customer_token:
        customer_token = rq.headers.get("X-Cb-CustomerToken", None)
        if customer_token:
            header_token = True

    return customer_token, header_token


customer_token_regex = re.compile(r"/CN=([-_A-Za-z0-9 ]+)(/|$)")
customer_token_regex2 = re.compile(r"/CN=([-_A-Za-z0-9 ]+)")


def parse_customer_token(header):
    if not header:
        return None

    match = re.search(customer_token_regex, header)
    if not match:
        return None

    if len(match.groups()) != 2:
        return None

    if len(re.findall(customer_token_regex2, header)) != 1:
        return None

    if match.groups()[1] not in {"/", ""}:
        return None

    return match.groups()[0]


def _get_request_id():
    request_id = None
    if flask.has_request_context():
        request_id = getattr(g, "request_id", None)
        if not request_id:
            request_id = request.headers.get("request_id")

            if not request_id:
                request_id = request.headers.get("X-Request-ID")

            g.request_id = request_id or str(uuid.uuid4())

    return request_id
