import os
import stat
from hashlib import sha256

import paramiko
from paramiko import PKey

PKey.get_fingerprint = lambda x: sha256(x.asbytes()).digest()  # monkeypatch for FIPS


class SFTPClient:
    """
    Simple wrapper around paramiko.SFTPClient
    """

    def __init__(self, host_name: str, user_name: str, password: str):
        self.host = host_name
        self.user_name = user_name
        self.password = password
        self._ssh_client = None
        self._sftp = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get_sftp_connection(self) -> paramiko.SFTPClient:
        if self._sftp is not None:
            if (
                self._sftp.get_channel()
                and self._sftp.get_channel().get_transport()
                and self._sftp.get_channel().get_transport().is_active()
            ):
                return self._sftp
            self._sftp.close()

        if self._ssh_client is None:
            self._ssh_client = paramiko.SSHClient()
            self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        if (
            not self._ssh_client.get_transport()
            or not self._ssh_client.get_transport().is_active()
        ):
            self._ssh_client.connect(
                hostname=self.host,
                username=self.user_name,
                password=self.password,
                timeout=600,
                look_for_keys=False,
            )
        self._sftp = self._ssh_client.open_sftp()
        print("reconnected")
        return self._sftp

    def upload_file(self, local_file_path, remote_file_path):

        for i in range(1, 4):
            try:
                sftp = self.get_sftp_connection()
                sftp.put(local_file_path, remote_file_path)
                print(f"copied {local_file_path} to {remote_file_path}")
                break
            except Exception as ex:
                # paramiko.ssh_exception.SSHException
                print(
                    f"try {i} Exception occurred while uploading {local_file_path}. {type(ex)}"
                )
                self.close()
                if i > 3:
                    raise

    def get_attrs(self, remote_path: str):
        try:
            sftp = self.get_sftp_connection()
            attrs = sftp.stat(remote_path)
            return attrs
        except FileNotFoundError:
            pass
        return None

    def close(self):
        if self._sftp:
            self._sftp.close()
            self._sftp = None
        if self._ssh_client:
            self._ssh_client.close()
            self._ssh_client = None

    def upload_folder(self, local_folder, remote_folder, preserve_mtime=True):

        for entry in os.listdir(local_folder):
            remote_path = remote_folder + "/" + entry
            attr = self.get_attrs(remote_path)
            if attr:
                print(f"{remote_path} exists")
                print(attr)
            local_path = os.path.join(local_folder, entry)
            if not os.path.isfile(local_path):
                if attr is None:
                    self.get_sftp_connection().mkdir(remote_path)
                self.upload_folder(local_path, remote_path, preserve_mtime)
            else:
                upload_file: bool = False
                if attr is None:
                    upload_file = True
                else:
                    local_file_attr = os.stat(local_path)
                    if attr.st_size != local_file_attr.st_size:
                        upload_file = True
                    if attr.st_mtime != local_file_attr.st_mtime:
                        upload_file = True
                if upload_file:
                    self.upload_file(local_path, remote_path)

    def remove_if_not_exists_locally(self, local_folder, remote_folder):
        sftp = self.get_sftp_connection()
        for entry in sftp.listdir_attr(remote_folder):
            remote_path = remote_folder + "/" + entry.filename
            local_path = local_folder + "/" + entry.filename
            if stat.S_ISDIR(entry.st_mode):
                self.remove_if_not_exists_locally(local_path, remote_path)
            else:
                if os.path.exists(local_path):
                    pass
                    # print(f"{local_path} exists")
                else:
                    sftp.remove(remote_path)
                    print(f"{remote_path} removed")

    def sync(self, local_folder: str, remote_folder: str):
        """
        :param local_folder: local folder that needs to be transferred
        :param remote_folder: remote location where the transfer will be stored
        """
        self.remove_if_not_exists_locally(local_folder, remote_folder)
        self.upload_folder(local_folder, remote_folder)

    def upload_file_obj(self, file_obj, remote_folder_path: str, remote_file_name: str):

        remote_file_path = remote_folder_path + "/" + remote_file_name
        for i in range(1, 4):
            try:
                sftp = self.get_sftp_connection()
                attr = self.get_attrs(remote_folder_path)
                if attr is None:
                    print(f"Created remote folder {remote_file_path}")
                    sftp.mkdir(remote_folder_path)
                print(f"Uploading the file to {remote_file_path}")
                sftp.putfo(file_obj, remote_file_path)
                print(f"Uploaded the file to {remote_file_path}")
                break
            except Exception as ex:
                # paramiko.ssh_exception.SSHException
                print(
                    f"try {i} Exception occurred while uploading the file. type={type(ex)}, ex={ex}"
                )
                self.close()
                if i >= 3:
                    raise
