from ast import literal_eval
from sqlalchemy.sql import text, and_
from typing import List

from tic.shared.ubs.models.ubs_analysis_models import (
    Main,
    Version,
    Status,
    Stage,
    MainPartnerLink,
    MainStageLink,
    StageStatus,
    StageTarget,
    StageTopic,
    Pipeline,
    MainPipelineLink,
    PipelineStatus,
    PipelineStageLink,
)

from tic.shared.ubs.util import hash_to_binary, binary_to_hash
from tic.shared.utils.date_util import utc_now
from tic.shared.ubs.dtos.analysis.pipeline_summary import PipelineSummary
from tic.shared.ubs.dtos.analysis.stage_summary import StageSummary
from tic.shared.ubs.enums.analysis.stage import UbsAnalysisStage
from tic.shared.ubs.enums.analysis.status import UbsAnalysisStatus
from tic.shared.ubs.enums.analysis.pipeline import UbsAnalysisPipeline
from tic.shared.ubs.repository.ubs_db_base import UbsDbRepositoryBase
from tic.shared.utils.hash_util import calculate_sha256


class UbsAnalysisDbRepository(UbsDbRepositoryBase):
    def __init__(
        self,
        logger,
        writer_conn_string,
        reader_conn_string,
        connection_args=None,
        statement_timeout=100,
    ):
        super(UbsAnalysisDbRepository, self).__init__(
            logger,
            writer_conn_string,
            reader_conn_string,
            connection_args,
            statement_timeout,
        )

    @staticmethod
    def update_stage_and_pipeline(
        sha256,
        stage: UbsAnalysisStage,
        stage_status: UbsAnalysisStatus,
        version,
        completion_date,
        writer,
    ):
        """
        @param sha256: hash to update
        @param stage: which stage had just finished
        @param stage_status: status of the stage that had just finished
        @param version: version of the stage that had just finished (major, minor)
        @param completion_date: completion_date if the hash has completed
        @param writer: ubs writer session for this transaction
        @return: an enum return of the status that the pipeline is now in
        """
        # TODO: placeholder until stage&pipeline SP is done
        writer.execute(
            text(
                "SELECT * FROM ubs_state.update_stage_pipeline_status(:in_sha256, (:in_stage_name)::varchar, (:in_stage_status)::varchar, (:in_stage_version_major)::smallint, (:in_stage_version_minor)::smallint, (:in_stage_completion_date)::timestamptz) LIMIT 1"
            ),
            {
                "in_sha256": hash_to_binary(sha256),
                "in_stage_name": stage.name,
                "in_stage_status": stage_status.name,
                "in_stage_version_major": version[0],
                "in_stage_version_minor": version[1],
                "in_stage_completion_date": completion_date,
            },
        )

    @staticmethod
    def update_stages_pipelines_get_targets(
        sha256,
        stage: UbsAnalysisStage,
        stage_status: UbsAnalysisStatus,
        version,
        completion_date,
        writer,
    ):
        """
        @param sha256: hash to fetch for
        @param stage: the stage that had just finished & that we want to fetch following stages of
        @param stage_status: the status of the stage that just updated
        @param version: version of the stage that we are looking at (major, minor)
        @param completion_date: date of when this stage was completed
        @param writer: ubs writer session for this transaction
        @return: a dict of {stages : arn}
        """
        next_stages = writer.execute(
            text(
                "SELECT * FROM ubs_state.update_stages_pipelines_get_targets(:in_sha256, (:in_stage_name)::varchar, (:in_stage_status)::varchar, (:in_version_major)::smallint, (:in_version_minor)::smallint, (:completion_date)::timestamptz)"
            ),
            {
                "in_sha256": hash_to_binary(sha256),
                "in_stage_name": stage.name,
                "in_stage_status": stage_status.name,
                "in_version_major": version[0],
                "in_version_minor": version[1],
                "completion_date": completion_date,
            },
        ).fetchall()

        ret = []
        for item in next_stages:
            if item[0] is None:
                return ret
            prep = literal_eval(item[0])
            listed = list(prep)
            for returned_stage in listed:
                stage_name, arn, requires_file = returned_stage[1:-1].split(",")
                ret.append(
                    {
                        "stage": UbsAnalysisStage[stage_name],
                        "arn": arn,
                        "requires_file": requires_file.lower() in {"t", "true"},
                    }
                )
        return ret

    @staticmethod
    def get_magic_and_file_details(sha256, reader) -> List[dict]:
        """
        @param sha256: the sha256 of the file
        @param reader: db reader session for this transaction
        @return: List of dict
        """
        result = reader.execute(
            text("SELECT * FROM ubs_state.get_magic_and_file_details(:in_sha256)"),
            {"in_sha256": hash_to_binary(sha256)},
        ).fetchall()

        return [
            {
                "sha256_truncated": binary_to_hash(item[0]),
                "s3_key": item[1],
                "truncated": item[2],
                "magic_value": item[3],
            }
            for item in result
        ]

    @staticmethod
    def update_magic_and_file_details(
        sha256, sha256_truncated, s3_key, truncated, magic_value, writer
    ):
        """
        @param sha256: the sha256 of the file
        @param writer: db writer session for this transaction
        @return: List of dict
        """
        magic_value_truncated = magic_value[:1024]

        writer.execute(
            text(
                "SELECT * FROM ubs_state.update_magic_and_file_details(:in_sha256, :in_sha256_truncated, (:in_s3_key)::varchar, (:in_truncated)::boolean, (:in_magic_value)::varchar, :in_magic_sha256)"
            ),
            {
                "in_sha256": hash_to_binary(sha256),
                "in_sha256_truncated": hash_to_binary(sha256_truncated),
                "in_s3_key": s3_key,
                "in_truncated": truncated,
                "in_magic_value": magic_value_truncated,
                "in_magic_sha256": hash_to_binary(
                    calculate_sha256(magic_value_truncated)
                ),
            },
        )

    def get_upload_history_for_file_and_partner(self, sha256, partner, reader):
        """
        Gets the upload history of a file to a given partner.
        @param sha256: the sha256 to check
        @param partner: A CbPartner object.
        @param reader: Reader session param, ubs_reader_session instance
        @return: The link if one exists, else None
        """

        return (
            reader.query(MainPartnerLink)
            .filter_by(partner_id=partner.value)
            .join(Main, Main.id == MainPartnerLink.main_id)
            .filter(Main.sha256 == hash_to_binary(sha256))
            .first()
        )

    def put_partner_upload(self, sha256, partner, partner_status, writer):
        """
        Puts a record of a partner upload attempt
        @param sha256: the file we've shared
        @param partner: the partner we've shared with (CbPartner expected)
        @param partner_status: the status the partner gave us (enum expected)
        @param writer: Writer session param, ubs_writer_session instance
        @return:
        """

        existing_link = self.get_upload_history_for_file_and_partner(
            sha256, partner, writer
        )
        if existing_link is not None:
            existing_link.response_state = partner_status.value
            existing_link.date_modified = utc_now()
            writer.add(existing_link)
        else:
            writer.add(
                MainPartnerLink(
                    partner_id=partner.value,
                    response_state=partner_status.value,
                    main=writer.query(Main)
                    .filter_by(sha256=hash_to_binary(sha256))
                    .first(),
                )
            )

    # ------------------------- FOR TESTING PURPOSES ONLY ------------------------- #
    @staticmethod
    def _insert_main(sha256, writer):
        """
        @param sha256: hash that you want to insert
        @param writer: ubs writer session for this transaction
        @return:
        """
        existing = (
            writer.query(Main).filter(Main.sha256 == hash_to_binary(sha256)).first()
        )
        if not existing:
            main = Main(sha256=hash_to_binary(sha256))
            writer.add(main)

    @staticmethod
    def _insert_version(major, minor, writer):
        """
        @param major: major version that we want to insert
        @param minor: minor version that we want to insert
        @param writer: ubs writer session for this transaction
        @return:
        """
        existing = (
            writer.query(Version)
            .filter(and_(Version.major == major, Version.minor == minor))
            .first()
        )
        if not existing:
            version = Version(major=1, minor=1)
            writer.add(version)

    @staticmethod
    def _insert_version_in(in_major, in_minor, writer):
        """
        @param major: major version that we want to insert
        @param minor: minor version that we want to insert
        @param writer: ubs writer session for this transaction
        @return:
        """
        existing = (
            writer.query(Version)
            .filter(and_(Version.major == in_major, Version.minor == in_minor))
            .first()
        )
        if not existing:
            version = Version(major=in_major, minor=in_minor)
            writer.add(version)

    @staticmethod
    def _insert_status(status: UbsAnalysisStatus, writer):
        """
        @param status: enum of ubsanalysisstate to insert
        @param writer: ubs writer session for this transaction
        @return:
        """
        existing = (writer.query(Status).filter(Status.name == status.name)).first()
        if not existing:
            status = Status(name=status.name)
            writer.add(status)

    @staticmethod
    def _insert_stage(stage: UbsAnalysisStage, major, minor, requires_file, writer):
        """
        @param stage: enum of ubsanalysisstage to insert
        @param major: current major version for the stage
        @param minor: current minor version for the stage
        @param writer: ubs writer session for this transaction
        @return:
        """
        version_id_result = (
            writer.query(Version.id)
            .filter(and_(Version.major == major, Version.minor == minor))
            .first()
        )
        existing = writer.query(Stage).filter(Stage.name == stage.name).first()
        if version_id_result:
            if not existing:
                stage = Stage(
                    name=stage.name,
                    current_version_id=version_id_result[0],
                    requires_file=requires_file,
                )
                writer.add(stage)

    @staticmethod
    def _update_pipeline(name, major, minor, writer):
        """
        @param name: name of the pipeline to update
        @param major: current major version for the pipeline
        @param minor: current minor version for the pipeline
        @param writer: ubs writer session for this transaction
        @return:
        """
        version_id = (
            writer.query(Version.id)
            .filter(and_(Version.major == major, Version.minor == minor))
            .first()
        )
        existing_pipeline = writer.query(Pipeline).filter(Pipeline.name == name).first()

        if version_id is None:
            raise ValueError(f"Version ({major}, {minor}) does not exist")

        version_id = version_id[0]

        if existing_pipeline is None:
            raise ValueError(f"Pipeline {name} does not exist")

        existing_pipeline.current_version_id = version_id
        writer.add(existing_pipeline)
        writer.flush()

    @staticmethod
    def _insert_pipeline(name, major, minor, writer):
        """
        @param name: name of the pipeline to insert
        @param major: current major version for the stage
        @param minor: current minor version for the stage
        @param writer: ubs writer session for this transaction
        @return:
        """
        version_id_result = (
            writer.query(Version.id)
            .filter(and_(Version.major == major, Version.minor == minor))
            .first()
        )
        existing = writer.query(Pipeline).filter(Pipeline.name == name).first()
        if version_id_result:
            if not existing:
                pipeline = Pipeline(name=name, current_version_id=version_id_result[0])
                writer.add(pipeline)

    @staticmethod
    def _insert_stage_target(
        src_stage: UbsAnalysisStage, target_stage: UbsAnalysisStage, writer
    ):
        """
        @param src_stage: src stage id (the dependency)
        @param target_stage: target stage id (the dependent)
        @param writer: ubs writer session for this transaction
        @return:
        """
        src_id_result = (
            writer.query(Stage.id).filter(Stage.name == src_stage.name).first()
        )
        target_id_result = (
            writer.query(Stage.id).filter(Stage.name == target_stage.name).first()
        )
        if src_id_result and target_id_result:
            existing = (
                writer.query(StageTarget)
                .filter(
                    and_(
                        StageTarget.stage_id == src_id_result[0],
                        StageTarget.target_stage_id == target_id_result[0],
                    )
                )
                .first()
            )
            if not existing:
                stage_target = StageTarget(
                    stage_id=src_id_result[0], target_stage_id=target_id_result[0]
                )
                writer.add(stage_target)

    @staticmethod
    def _insert_stage_topic(stage: UbsAnalysisStage, topic_arn, enabled, writer):
        """
        @param stage: stage to attach arn to
        @param topic_arn: sns arn to match to stage
        @param enabled: is this sns topic enabled
        @param writer: ubs writer session for this transaction
        @return:
        """
        result = writer.query(Stage.id).filter(Stage.name == stage.name).first()
        if result:
            stage_id = result[0]
            existing = (
                writer.query(StageTopic).filter(StageTopic.stage_id == stage_id).first()
            )
            if not existing:
                stage_topic = StageTopic(
                    stage_id=stage_id, sns_arn=topic_arn, enabled=enabled
                )
                writer.add(stage_topic)

    @staticmethod
    def _get_stage_status(in_sha256, in_stage, in_major, in_minor, in_session):
        query = (
            in_session.query(Status.name)
            .join(StageStatus, Status.id == StageStatus.status_id)
            .join(Version, Version.id == StageStatus.version_id)
            .filter(Version.major == in_major, Version.minor == in_minor)
            .join(MainStageLink, MainStageLink.id == StageStatus.main_stage_link_id)
            .join(Main, Main.id == MainStageLink.main_id)
            .filter(Main.sha256 == hash_to_binary(in_sha256))
            .join(Stage, Stage.id == MainStageLink.stage_id)
            .filter(Stage.name == in_stage.name)
            .first()
        )

        return query[0] if query else None

    @staticmethod
    def _get_pipeline_status(in_sha256, in_pipeline, in_major, in_minor, in_session):
        return (
            in_session.query(Status.name)
            .join(PipelineStatus, Status.id == PipelineStatus.status_id)
            .join(Version, Version.id == PipelineStatus.version_id)
            .filter(Version.major == in_major, Version.minor == in_minor)
            .join(
                MainPipelineLink,
                MainPipelineLink.id == PipelineStatus.main_pipeline_link_id,
            )
            .join(Main, Main.id == MainPipelineLink.main_id)
            .filter(Main.sha256 == hash_to_binary(in_sha256))
            .join(Pipeline, Pipeline.id == MainPipelineLink.pipeline_id)
            .filter(Pipeline.name == in_pipeline.name)
            .first()
        )[0]

    @staticmethod
    def _insert_main_stage_link(sha256, stage: UbsAnalysisStage, writer):
        """
        @param sha256: hash to link to stage
        @param stage: stage to link to hash
        @param writer: ubs writer session for this transaction
        @return:
        """
        stage_id_result = (
            writer.query(Stage.id).filter(Stage.name == stage.name).first()
        )
        main_id_result = (
            writer.query(Main.id).filter(Main.sha256 == hash_to_binary(sha256)).first()
        )
        if stage_id_result and main_id_result:
            existing = (
                writer.query(MainStageLink)
                .filter(
                    and_(
                        MainStageLink.main_id == main_id_result[0],
                        MainStageLink.stage_id == stage_id_result[0],
                    )
                )
                .first()
            )
            if not existing:
                main_stage_link = MainStageLink(
                    main_id=main_id_result[0], stage_id=stage_id_result[0]
                )
                writer.add(main_stage_link)

    @staticmethod
    def _insert_main_pipeline_link(sha256, pipeline_name, writer):
        """
        @param sha256: hash to link to pipeline
        @param pipeline_name: pipeline to link to hash
        @param writer: ubs writer session for this transaction
        @return:
        """
        pipeline_id_result = (
            writer.query(Pipeline.id).filter(Pipeline.name == pipeline_name).first()
        )
        main_id_result = (
            writer.query(Main.id).filter(Main.sha256 == hash_to_binary(sha256)).first()
        )
        if pipeline_id_result and main_id_result:
            existing = (
                writer.query(MainPipelineLink)
                .filter(
                    and_(
                        MainPipelineLink.main_id == main_id_result[0],
                        MainPipelineLink.pipeline_id == pipeline_id_result[0],
                    )
                )
                .first()
            )
            if not existing:
                main_pipeline_link = MainPipelineLink(
                    main_id=main_id_result[0], pipeline_id=pipeline_id_result[0]
                )
                writer.add(main_pipeline_link)

    @staticmethod
    def _insert_pipeline_stage_link(pipeline_name, stage: UbsAnalysisStage, writer):
        """
        @param pipeline_name: pipeline to link to stage
        @param stage: stage to link to pipeline
        @param writer: ubs writer session for this transaction
        @return:
        """
        stage_id_result = (
            writer.query(Stage.id).filter(Stage.name == stage.name).first()
        )
        pipeline_id_result = (
            writer.query(Pipeline.id).filter(Pipeline.name == pipeline_name).first()
        )
        if pipeline_id_result and stage_id_result:
            existing = (
                writer.query(PipelineStageLink)
                .filter(
                    and_(
                        PipelineStageLink.pipeline_id == pipeline_id_result[0],
                        PipelineStageLink.stage_id == stage_id_result[0],
                    )
                )
                .first()
            )
            if not existing:
                pipeline_stage_link = PipelineStageLink(
                    pipeline_id=pipeline_id_result[0], stage_id=stage_id_result[0]
                )
                writer.add(pipeline_stage_link)

    @staticmethod
    def _insert_pipeline_status(
        sha256, pipeline_name, major, minor, status: UbsAnalysisStatus, writer
    ):
        """
        @param sha256: hash for this current pipeline status
        @param pipeline_name: pipeline to update status of
        @param major: current major version for the stage
        @param minor: current minor version for the stage
        @param status: status to update pipeline to
        @param writer: ubs writer session for this transaction
        @return:
        """
        main_id_result = (
            writer.query(Main.id).filter(Main.sha256 == hash_to_binary(sha256)).first()
        )
        pipeline_id_result = (
            writer.query(Pipeline.id).filter(Pipeline.name == pipeline_name).first()
        )
        mpl_id_result = (
            writer.query(MainPipelineLink.id)
            .filter(
                and_(
                    MainPipelineLink.main_id == main_id_result[0],
                    MainPipelineLink.pipeline_id == pipeline_id_result[0],
                )
            )
            .first()
        )
        status_id_result = (
            writer.query(Status.id).filter(Status.name == status.name).first()
        )
        version_id_result = (
            writer.query(Version.id)
            .filter(and_(Version.major == major, Version.minor == minor))
            .first()
        )

        if mpl_id_result and status_id_result and version_id_result:
            pipeline_status = PipelineStatus(
                main_pipeline_link_id=mpl_id_result[0],
                status_id=status_id_result[0],
                version_id=version_id_result[0],
            )
            writer.add(pipeline_status)

    @staticmethod
    def _insert_stage_status(
        sha256, stage: UbsAnalysisStage, version, status: UbsAnalysisStatus, writer
    ):
        """
        @param sha256: hash for this current pipeline status
        @param stage: stage to update status of
        @param version: current major version for the stage
        @param status: status to update pipeline to
        @param writer: ubs writer session for this transaction
        @return:
        """
        main_id_result = (
            writer.query(Main.id).filter(Main.sha256 == hash_to_binary(sha256)).first()
        )
        stage_id_result = (
            writer.query(Stage.id).filter(Stage.name == stage.name).first()
        )
        msl_id_result = (
            writer.query(MainStageLink.id)
            .filter(
                and_(
                    MainStageLink.main_id == main_id_result[0],
                    MainStageLink.stage_id == stage_id_result[0],
                )
            )
            .first()
        )
        status_id_result = (
            writer.query(Status.id).filter(Status.name == status.name).first()
        )
        version_id_result = (
            writer.query(Version.id)
            .filter(and_(Version.major == version[0], Version.minor == version[1]))
            .first()
        )

        if msl_id_result and status_id_result and version_id_result:
            stage_status = StageStatus(
                main_stage_link_id=msl_id_result[0],
                status_id=status_id_result[0],
                version_id=version_id_result[0],
            )
            writer.add(stage_status)

    @staticmethod
    def start_pipeline(
        writer, sha256, pipeline: UbsAnalysisPipeline, force_restart=False
    ) -> PipelineSummary:
        """
        @param sha256: File for which to run the pipeline
        @param pipeline: the pipeline to start if needed. Status of this pipeline and its stages will be returned.
        @param force_restart: Flag to restart pipeline execution by force if it's already in progress or completed.
        @param writer: ubs writer session for this transaction
        @return: TODO
        """
        pipeline_status = writer.execute(
            text(
                "SELECT * FROM ubs_state.start_pipeline(:in_sha256, (:in_pipeline_name)::varchar, (:force_restart)::bool)"
            ),
            {
                "in_sha256": hash_to_binary(sha256),
                "in_pipeline_name": pipeline.name,
                "force_restart": force_restart,
            },
        ).fetchone()

        pipeline_summary = PipelineSummary(sha256)
        pipeline_summary.name = UbsAnalysisPipeline[pipeline_status[1]]
        pipeline_summary.status = UbsAnalysisStatus[pipeline_status[2]]
        pipeline_summary.version = tuple(pipeline_status[3])
        # custom types and arrays are a pain.
        for raw_stage in list(literal_eval(pipeline_status[4])):
            stage_parts = raw_stage[1:-1].split(",", 2)
            pipeline_summary.stages.append(
                StageSummary(
                    name=UbsAnalysisStage[stage_parts[0]],
                    status=UbsAnalysisStatus[stage_parts[1]],
                    version=tuple([int(x) for x in stage_parts[2][2:-2].split(",")]),
                )
            )
        return pipeline_summary

    @staticmethod
    def get_pipeline_executable_root_stages(
        reader,
        sha256,
        pipeline: UbsAnalysisPipeline,
    ) -> List[dict]:
        """
        @param reader: UBS reader session for this transaction
        @param sha256: the sha256 of the file for which we may need to run stages
        @param pipeline: the pipeline to find the next stages to run
        @return: List of dict {stage: sns_arn}
        """

        stage_targets = reader.execute(
            text(
                "SELECT * FROM ubs_state.get_pipeline_executable_root_stages(:in_sha256, (:in_pipeline_name)::varchar)"
            ),
            {"in_sha256": hash_to_binary(sha256), "in_pipeline_name": pipeline.name},
        ).fetchall()

        return [
            {
                "stage": UbsAnalysisStage[item[0]],
                "sns_arn": item[1],
                "version": tuple(item[2]),
            }
            for item in stage_targets
        ]

    @staticmethod
    def get_current_stage_statuses(reader, sha256) -> List[dict]:
        """
        Get the statuses of all stages, that have a status, for the current version.

        @param reader: UBS reader session for this transaction
        @param sha256: The sha256 to get the status for
        @return: List of dict {stage, status}
        """

        stage_statuses = (
            reader.query(Stage.name.label("stage"), Status.name.label("status"))
            .select_from(Main)
            .filter(Main.sha256 == hash_to_binary(sha256))
            .join(MainStageLink, Main.id == MainStageLink.main_id)
            .join(Stage, MainStageLink.stage_id == Stage.id)
            .join(
                StageStatus,
                (MainStageLink.id == StageStatus.main_stage_link_id)
                & (Stage.current_version_id == StageStatus.version_id),
            )
            .join(Status, StageStatus.status_id == Status.id)
            .all()
        )

        # Not actually a private method - underscore is to allow compatibility with namedtuple
        # See: https://docs.sqlalchemy.org/en/13/orm/query.html#sqlalchemy.util.KeyedTuple._asdict
        # noinspection PyProtectedMember
        return [x._asdict() for x in stage_statuses]

    def prune(self):
        """
        Prunes ubs state data store. Mostly keep recent data stage_status and main_stage_link and delete old data
        """
        with self.ubs_writer_session() as writer:
            writer.execute(
                text("call ubs_state.prune();").execution_options(autocommit=True)
            )
