"""
Validation functionality for IPv4 IOCs.Z

This is intended to validate lists of IOCs using a built in cleanlist and check if the IP Address is public.

In the future - the cleanlist should be moved to a database where we

Can receive a logging object and will use. Otherwise will set up a standard
logger to go to STDERR/STDOUT. Pass skip to have no logs.
"""

from tic.shared.utils.parsers import network
from .validation_base import Validation, pass_score_threshold


class ValidateIPv4(Validation):
    # List of valid network types
    _valid_ipv4_types = [network.AuthorityType.ipv4]

    def __init__(
        self,
        input_data=None,
        validation_types=("ipv4_cleanlist", "is_public_ipv4"),
        score_threshold=0,
        **kwargs
    ):
        """
        :param validation_types: Initialize with this validation type.
        :param score_threshold: Not really used here yet - just True/False Passing
        :param kwargs: Other arguments passed to ValidationBase
        :return:
        """
        if input_data is None:
            input_data = list()
        super(ValidateIPv4, self).__init__(**kwargs)

        self._ioc_type = "ipv4"

        self.set_input_data(input_data)

        for validation_type in validation_types:
            if validation_type == "ipv4_cleanlist":
                self.add_validation_method("ipv4_cleanlist", check_cleanlist)
                self.add_score_threshold_method(
                    "ipv4_cleanlist", pass_score_threshold, score_threshold
                )
            elif validation_type == "is_public_ipv4":
                self.add_validation_method("is_public_ipv4", check_is_public_ipv4)
                self.add_score_threshold_method(
                    "is_public_ipv4", pass_score_threshold, score_threshold
                )
            else:
                raise ValueError("Invalid validation type {0}".format(validation_type))

    def set_input_data(self, input_ipv4):
        """
        Will raise an Exception if any of the elements are invalid when adding the input data.

        :param input_ipv4: List of IPv4 addresses.
        :return: Will set an internal variable of bad_iocs.
        """
        for i in input_ipv4:
            network_type, temp_network = network.determine_authority_type(i)
            if network_type not in self._valid_ipv4_types:
                self.bad_input_data.append(i)
            else:
                self._input_data.append(i)

        if self.bad_input_data:
            raise ValueError("ValidateIPv4: Invalid IOCs Inputted")


def check_cleanlist(input_ipv4, results):
    """
    This should eventually become a database and cleanlist to check - with the address being represented as ranges that
    we chose to trust.

    :param input_ipv4: List of IPv4 Addresses to check
    :return:
    """
    try:
        for ipv4 in input_ipv4:
            if ipv4 in IPv4_CLEANLIST:
                continue
            else:
                results.append(ipv4)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


def check_is_public_ipv4(input_ipv4, results):
    """
    :param input_ipv4: List of IPv4 Addresses to check that are Public - performing a strict check.
    """
    try:
        for ipv4 in input_ipv4:
            if network.is_public_ipv4(ipv4):
                results.append(ipv4)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


IPv4_CLEANLIST = [
    # Google DNS Servers
    "8.8.8.8",
    "8.8.4.4",
    # Some CarbonBlack IPs
    "54.86.237.241",  # carbonblack.com
    "52.21.5.219",  # threatintel.bit9.com
    "54.236.137.222",  # threatintel.bit9.com
]
