"""
Validation functionality for IPv6 IOCs.Z

This is intended to validate lists of IOCs using a built in cleanlist and check if the IP Address is public.

In the future - the cleanlist should be moved to a database where we

Can receive a logging object and will use. Otherwise will set up a standard
logger to go to STDERR/STDOUT. Pass skip to have no logs.
"""


from tic.shared.utils.parsers import network
from netaddr import IPNetwork

from .validation_base import pass_score_threshold, Validation


class ValidateIPv6(Validation):
    # List of valid network types
    _valid_ipv6_types = [network.AuthorityType.ipv6, network.AuthorityType.ipv6_subnet]

    def __init__(
        self,
        input_data=None,
        validation_types=("ipv6_cleanlist", "is_public_ipv6", "check_prefix_len"),
        score_threshold=0,
        **kwargs
    ):
        """
        :param validation_types: Initialize with this validation type.
        :param score_threshold: Not really used here yet - just True/False Passing
        :param kwargs: Other arguments passed to ValidationBase
        :return:
        """
        if input_data is None:
            input_data = list()
        super(ValidateIPv6, self).__init__(**kwargs)

        self._ioc_type = "ipv6"

        self.set_input_data(input_data)

        for validation_type in validation_types:
            if validation_type == "ipv6_cleanlist":
                self.add_validation_method("ipv6_cleanlist", check_cleanlist)
                self.add_score_threshold_method(
                    "ipv6_cleanlist", pass_score_threshold, score_threshold
                )
            elif validation_type == "is_public_ipv6":
                self.add_validation_method("is_public_ipv6", check_is_public_ipv6)
                self.add_score_threshold_method(
                    "is_public_ipv6", pass_score_threshold, score_threshold
                )
            elif validation_type == "check_prefix_len":
                self.add_validation_method("check_prefix_len", check_prefix_len)
                self.add_score_threshold_method(
                    "check_prefix_len", pass_score_threshold, score_threshold
                )
            else:
                raise ValueError("Invalid validation type {0}".format(validation_type))

    def set_input_data(self, input_ipv6):
        """
        Will raise an Exception if any of the elements are invalid when adding the input data.

        :param input_ipv6: List of IPv6 addresses.
        :return: Will set an internal variable of bad_iocs.
        """
        for i in input_ipv6:
            network_type, temp_network = network.determine_authority_type(i)
            if network_type not in self._valid_ipv6_types:
                self.bad_input_data.append(i)
            else:
                self._input_data.append(i)

        if self.bad_input_data:
            raise ValueError("ValidateIPv6: Invalid IOCs Inputted")


def check_cleanlist(input_ipv6, results):
    """
    This should eventually become a database and cleanlist to check - with the address being represented as ranges that
    we chose to trust.

    :param input_ipv6: List of IPv6 Addresses to check
    :return:
    """
    try:
        for ipv6 in input_ipv6:
            input_network = IPNetwork(ipv6)

            if not _check_ipv6_cleanlist(input_network):
                results.append(ipv6)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


def _check_ipv6_cleanlist(input_network):
    for cleanlist_network in IPv6_CLEANLIST:
        if input_network in cleanlist_network:
            return True


def check_prefix_len(input_ipv6, results):
    """
    :param input_ipv6: List of IPv6 Networks to check that the prefix length of a network isn't too large.
    """
    MAX_PREFIX_LEN = 64

    try:
        for ipv6 in input_ipv6:
            if IPNetwork(ipv6).prefixlen >= MAX_PREFIX_LEN:
                results.append(ipv6)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


def check_is_public_ipv6(input_ipv6, results):
    """
    :param input_ipv6: List of IPv6 Addresses to check that are Public - performing a strict check.
    """
    try:
        for ipv6 in input_ipv6:
            if network.is_public_ipv6(ipv6):
                results.append(ipv6)

        return results
    except Exception as e:
        # Delete all elements in the list
        del results[:]

        # Set the list to one element that is the exception for this thread.
        results.append(e)
        return None


# These are public IPv6 addresses or networks that are cleanlisted for whatever reason.
IPv6_CLEANLIST = (
    # Google DNS Servers
    IPNetwork("2001:4860:4860::8888"),
    IPNetwork("2001:4860:4860::8844"),
    # YouTube Range
    IPNetwork("2620:11a:a000::/40"),
)
