"""
Base Class for all IOC specific validation Classes for IOC Validation.

Can receive a logging object and will use. Otherwise will set up a standard
logger to go to STDERR/STDOUT. Pass skip to have no logs.

"""

import logging
import threading
import time


class Validation(object):
    """
    Validation Class forms the base of Validation for IOCs.

    This is not intended to be used directly but inherited by IOC specific child classes.

    Internal Variables -
        _bad_data             - Input data that falls outside of the scoring threshold or another test.
        _input_data           - The raw list of inputted IOCs - set only once.
        _validated_data       - The IOCs that met the score threshold
        _scoring_thresholds   - Scoring thresholds that are set for validation methods.

    Internal Config
        _ioc_type          - The IOC type for the instance.
        _form_regex_match  - Regexps to filter out data based on a .match
        _form_regex_search - Regexps to filter out data based on a .search

        _form_compiled_regex_match  - re.compiled from _form_regex_match
        _form_compiled_regex_search - re.compiled from _form_regex_search
    """

    # Standard class instance attributes
    _internal = ["_input_data", "_validated_data"]

    # Standard class instance configuration
    _internal_config = ["_ioc_type"]

    def __init__(self, logger=None):
        """
        :param logger: Pass "skip" to have 0 logging or pass a logging.Logger instance.
        :return:
        """
        if logger is None:
            # Set up a default logger to print to STDOUT if no logger is provided
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(logging.INFO)

            sh = logging.StreamHandler()

            formatter = logging.Formatter()
            sh.setFormatter(formatter)

            self.logger.addHandler(sh)
        elif logger == "skip":
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.NullHandler())
            self.logger.setLevel(0)
        else:
            self.logger = logger

        self._input_data = []

        # List that will hold any bad IOCs when _input_data is set.
        self.bad_input_data = []

        # self._well_formed_data = []
        # self._not_well_formed_data = []

        self._ioc_type = None

        self._validation_source = {}
        self._score_thresholds = {}
        self._score_thresholds_check = {}

        # Methods with which we call external services.
        self._validation_methods = {}

        # Extra arguments to be passed to those as **args
        self._validation_methods_extra_args = {}

        # Extra arguments to be passed to those as **args
        self._score_thresholds_methods_extra_args = {}

        # Responses from the validation methods
        self._validation_method_responses = {}

        # Results after applying the score threshold methods
        self._scoring_methods_results = {}

        # Valid score ranges for a validation method
        self._validation_score_ranges = {}

        self._validated_data = []

        self._false_positives = []

        self.logger.debug("ioc_validation: Validation Base Class initialized")

    def get_validated_data(self):
        if not self._validated_data:
            self.call_all_validation_methods()
            self.get_iocs_above_thresholds()

        return self._validated_data

    def get_false_positives(self):
        self.get_validated_data()

        return self._false_positives

    def log_current_config(self):
        """
        Emit a log message with the current configuration of the instance.
        """

        self.logger.info(
            "ioc_validation: is currently configured - {0}".format(
                self.config_to_dict()
            )
        )

    def config_to_dict(self):
        """
        :return: A dictionary with the current configuration of the instance.
        """
        d = {}

        for k in self._internal_config:
            d[k] = self.__getattribute__(k)

        # Include the logger name and level in the configuration output as well.
        d["logger.name"] = self.logger.name
        d["logger.level"] = self.logger.level

        return d

    def add_score_threshold_method(
        self, validation_method, threshold_function, score_threshold, extra_args=None
    ):
        """
        :param validation_method: String that represents the validation method, example "srs"
        :param threshold_function: Function that will parse the data and return results based on score_threshold
        :param score_threshold: Integer representing the score that all IOCs need to be above to be valid
        :param extra_args: Extra arguments to be passed to the threshold_function
        :return:
        """
        if extra_args is None:
            extra_args = dict()
        # If we know the valid score ranges for a score threshold method - check it.
        if validation_method in self._validation_score_ranges:
            if (
                score_threshold < self._validation_score_ranges[validation_method][0]
                or score_threshold > self._validation_score_ranges[validation_method][1]
            ):
                raise ValueError("add_score_threshold_method: Invalid score_threshold")

        self._score_thresholds[validation_method] = score_threshold
        self._score_thresholds_check[validation_method] = threshold_function
        self._score_thresholds_methods_extra_args[validation_method] = extra_args

    def add_validation_method(
        self, validation_method, validation_function, extra_args=None
    ):
        """
        :param validation_method: String that represents name of the validation method, example "srs"
        :param validation_function: Validation function itself that will make the call given an list of IOCs
        :param extra_args: Extra arguments to pass to the validation method
        :return:
        """
        if extra_args is None:
            extra_args = dict()
        self.logger.debug(
            "add_validation_method: added validation_method {0} function name {1} extra args {2}".format(
                validation_method, validation_function.__name__, extra_args
            )
        )

        self._validation_methods[validation_method] = validation_function
        self._validation_methods_extra_args[validation_method] = extra_args

    def call_all_validation_methods(self):
        """
        Call all validation methods that have been set.

        """

        time_start = time.time()

        threads = {}
        thread_results = {}

        self.logger.debug(
            "call_all_validation_methods: start for {0} validation methods".format(
                len(self._validation_methods)
            )
        )

        for key in self._validation_methods.keys():
            self.logger.debug(
                "call_all_validation_methods: Calling validation method {0} for {1} iocs".format(
                    key, len(self._input_data)
                )
            )

            extra_args = self._validation_methods_extra_args[key]
            thread_results[key] = []
            t = threading.Thread(
                target=self._validation_methods[key],
                args=(self._input_data, thread_results[key]),
                kwargs=extra_args,
            )
            threads[key] = t
            t.start()

        for key in threads:
            threads[key].join()
            self._validation_method_responses[key] = thread_results[key]

            # Signal an error retrieving results in the thread, pass the exception up out of it.
            if self._validation_method_responses[key] and isinstance(
                thread_results[key][0], Exception
            ):
                self.logger.warning(
                    "call_all_validation_methods: Exception in calling validation method {0}: {1}".format(
                        self._validation_methods[key].__name__, thread_results[key][0]
                    )
                )
                raise thread_results[key][0]

            self.logger.debug(
                "call_all_validation_methods: Finished calling validation method {0} got {1} iocs returned".format(
                    key, len(self._validation_method_responses[key])
                )
            )

        self.logger.debug(
            "call_all_validation_methods: finished for {0} validation methods - took {1} seconds".format(
                len(self._validation_methods), round(time.time() - time_start, 3)
            )
        )

    def get_iocs_above_thresholds(self):
        """
        Return list of IOCs above a certain threshold after processing self._score_thresholds with
        self._score_thresholds for all validation_methods.

        IOCs are also put into _validated_data
        """
        self.logger.debug(
            "get_iocs_above_thresholds: starting to filter through results of {0} iocs".format(
                len(self._input_data)
            )
        )

        for validation_method in self._score_thresholds:
            score_threshold = self._score_thresholds[validation_method]
            score_method = self._score_thresholds_check[validation_method]
            raw_response = self._validation_method_responses[validation_method]
            extra_args = self._score_thresholds_methods_extra_args.get(
                validation_method, {}
            )
            self._scoring_methods_results[validation_method] = set(
                score_method(raw_response, score_threshold, **extra_args)
            )
        # Only include the indicators that all validation methods agree on
        self._merge_results()

        # False positives will be (inputted results - the validated results)
        self._false_positives = list(set(self._input_data) - set(self._validated_data))

    def _merge_results(self):
        results = []

        for validation_method in self._scoring_methods_results:
            results.append(set(self._scoring_methods_results[validation_method]))

        temp_set = set.intersection(*results)

        num_filtered = len(self._input_data) - len(temp_set)
        if num_filtered > 0:
            self.logger.warning(
                f"get_iocs_above_thresholds: filtered out {num_filtered} iocs"
            )

        # Remove duplicates
        self._validated_data = list(temp_set)


def pass_score_threshold(results, _):
    """
    "Pass" function for score threshold - use when there isn't a score threshold to check.

    """
    for r in results:
        yield r
